@@ -25,12 +25,9 @@
  * SPDX-License-Identifier: MIT
  */
 
-#include "ares_setup.h"
-
-#include <assert.h>
-
-#include "ares.h"
 #include "ares_private.h"
+#include "ares_event.h"
+#include <assert.h>
 
 void ares_destroy(ares_channel_t *channel)
 {
@@ -41,6 +38,37 @@ void ares_destroy(ares_channel_t *channel)
     return;
   }
 
+  /* Mark as being shutdown */
+  ares__channel_lock(channel);
+  channel->sys_up = ARES_FALSE;
+  ares__channel_unlock(channel);
+
+  /* Disable configuration change monitoring.  We can't hold a lock because
+   * some cleanup routines, such as on Windows, are synchronous operations.
+   * What we've observed is a system config change event was triggered right
+   * at shutdown time and it tries to take the channel lock and the destruction
+   * waits for that event to complete before it continues so we get a channel
+   * lock deadlock at shutdown if we hold a lock during this process. */
+  if (channel->optmask & ARES_OPT_EVENT_THREAD) {
+    ares_event_thread_t *e = channel->sock_state_cb_data;
+    if (e && e->configchg) {
+      ares_event_configchg_destroy(e->configchg);
+      e->configchg = NULL;
+    }
+  }
+
+  /* Wait for reinit thread to exit if there was one pending, can't be
+   * holding a lock as the thread may take locks. */
+  if (channel->reinit_thread != NULL) {
+    void *rv;
+    ares__thread_join(channel->reinit_thread, &rv);
+    channel->reinit_thread = NULL;
+  }
+
+  /* Lock because callbacks will be triggered, and any system-generated
+   * callbacks need to hold a channel lock. */
+  ares__channel_lock(channel);
+
   /* Destroy all queries */
   node = ares__llist_node_first(channel->all_queries);
   while (node != NULL) {
@@ -48,12 +76,14 @@ void ares_destroy(ares_channel_t *channel)
     struct query       *query = ares__llist_node_claim(node);
 
     query->node_all_queries = NULL;
-    query->callback(query->arg, ARES_EDESTRUCTION, 0, NULL, 0);
+    query->callback(query->arg, ARES_EDESTRUCTION, 0, NULL);
     ares__free_query(query);
 
     node = next;
   }
 
+  ares_queue_notify_empty(channel);
+
 #ifndef NDEBUG
   /* Freeing the query should remove it from all the lists in which it sits,
    * so all query lists should be empty now.
@@ -69,6 +99,14 @@ void ares_destroy(ares_channel_t *channel)
   assert(ares__htable_asvp_num_keys(channel->connnode_by_socket) == 0);
 #endif
 
+  /* No more callbacks will be triggered after this point, unlock */
+  ares__channel_unlock(channel);
+
+  /* Shut down the event thread */
+  if (channel->optmask & ARES_OPT_EVENT_THREAD) {
+    ares_event_thread_destroy(channel);
+  }
+
   if (channel->domains) {
     for (i = 0; i < channel->ndomains; i++) {
       ares_free(channel->domains[i]);
@@ -89,13 +127,17 @@ void ares_destroy(ares_channel_t *channel)
 
   ares__hosts_file_destroy(channel->hf);
 
+  ares__qcache_destroy(channel->qcache);
+
+  ares__channel_threading_destroy(channel);
+
   ares_free(channel);
 }
 
 void ares__destroy_server(struct server_state *server)
 {
   if (server == NULL) {
-    return;
+    return; /* LCOV_EXCL_LINE: DefensiveCoding */
   }
 
   ares__close_sockets(server);
